/*
 * Copyright (c) 2018 Alibaba Group Holding Ltd. All rights reserved.
 */

'use strict';

const Log = require('../../../utils/log');
const Group = require('../group');
const SystemFunction = require('./index');
const constant = require('../../../utils/constant');
const dbusProxy = require('../../dbus/dbus-proxy');

const TAG = '[Function-Thing]';
const DBUS_PATTERN = /^[\w-@:\.]+$/;

class Thing extends SystemFunction {

  constructor(functionService) {
    super(functionService);
    this.id = constant.THING_FUNCTION_ID;
  }

  handler(event, context, callback) {
    Log.i(TAG, `Handler runs with "${event}", "${JSON.stringify(context)}".`);
    var topic = context.invokerContext.custom.topic;
    if (!topic) {
      callback(new Error('Topic is not found in InvokerContext.'));
      return;
    }
    if (event === undefined || event === null) {
      callback(new Error(`Payload is required: "${event}".`));
      return;
    }
    var args;
    try {
      args = JSON.parse(event.toString());
    } catch (error) {
      callback(new Error('Payload is not JSON compatible.'));
      return;
    }
    this._handler(topic, args, callback);
  }

  _handler(topic, args, callback) {
    var splits = topic.split('/');
    // TODO Check the topic's format.
    var service = splits[6];
    if (service === undefined || service === null) {
      callback(new Error('Custom.topic in InvokerContext is undefined or null.'));
      return;
    }
    var productKey = splits[3];
    var deviceName = splits[4];
    if (productKey && deviceName) {
      this._handleCallThingService(productKey, deviceName, service, args, callback);
      return;
    }
    if (service === 'getthingswithtags') {
      this._handleGetThingWithTags(args, callback);
      return;
    }
    callback(new Error(`Unknown request "${service}" with args "${args}".`));
  }

  _handleCallThingService(productKey, deviceName, service, args, callback) {
    if (!productKey || !DBUS_PATTERN.test(productKey)) {
      callback(new Error(`ProductKey ${productKey} is illegal. Allowed characters are [a-zA-Z0-9_-@:.].`));
      return;
    }
    if (!deviceName || !DBUS_PATTERN.test(deviceName)) {
      callback(new Error(`DeviceName ${deviceName} is illegal. Allowed characters are [a-zA-Z0-9_-@:.].`));
      return;
    }
    if (service === undefined || service === null) {
      callback(new Error(`Service ${service} is illegal.`));
      return;
    }
    var id = productKey + '_' + deviceName;
    if (service === 'get') {
      dbusProxy.getBatchPropertiesAsync(id, args, callback);
    } else if (service === 'set') {
      dbusProxy.setBatchPropertiesAsync(id, args, callback);
    } else {
      dbusProxy.callServiceAsync(id, service, args, callback);
    }
  }

  _handleGetThingWithTags(thingTags, callback) {
    if (!thingTags || !Array.isArray(thingTags)) {
      callback(new Error('Payload is not a array.'));
      return;
    }
    var group = new Group();
    group.getThingsWithTags(thingTags, function (err, things) {
      if (err) {
        callback(err);
      } else {
        callback(null, JSON.stringify(things));
      }
    });
  }
}

module.exports = Thing;